import { MainWrapper } from "../components/MainWrapper";

export const decorators = [Story => <MainWrapper>{<Story />}</MainWrapper>];

export const parameters = {
  actions: { argTypesRegex: "^on[A-Z].*" },
  controls: {
    matchers: {
      color: /(background|color)$/i,
      date: /Date$/,
    },
  },
};
